/* Password matrix generator

   Original Idea from John Graham-Cumming
   http://blog.jgc.org/2010/12/write-your-passwords-down.html

   Usage:  passgen <random_file

   The random file must come from a cryptographically secure
   random source (typically /dev/random).
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

// Just a debug utility
void check_range(int min, int max, int val, char* name)
{
  if (val < min || val > max) {
    fprintf(stderr, "Programming error: %s = %d out of range.", name, val);
    exit(1);
  }
}

// Return 2^(nb_ones + 1) - 1. For instance, ones(3) == 0b111
// That makes handy bit masks
int ones(int nb_ones)
{
  check_range(0, 8, nb_ones, "nb_ones");
  return (1 << nb_ones) - 1;
}

// Gives random bits, in an integer.  They are placed
// in the least significant positions.
// We shouldn't ask for more than 8 random bits.
int get_bits(int nb_bits)
{
  // We start with no random bit
  static int buf = 0;
  static int bufsize = 0;

  check_range(0, 8, nb_bits, "nb_bits");

  // get random bits from stdin if needed
  if(nb_bits > bufsize) {
    buf     += getchar()  << bufsize;
    bufsize += 8;
  }

  // take random bits from buffer
  int result = buf & ones(nb_bits);

  // remove the bits we've just taken
  buf     &= ~ones(nb_bits);
  buf    >>= nb_bits;
  bufsize -= nb_bits;

  return result;
}

// Gives a random int, ranging from 0 to range - 1
int get_range(int range)
{
  check_range(1, 256, range, "range");
  int nb_digits = 0;
  int temp      = range - 1;
  for (;temp != 0; temp /= 2)
    nb_digits ++;

  return get_bits(nb_digits) % range;
}

// Prints a square that will contain your passwords
// The square will be adequately decorated, with
// letter if start == 'A', or digits if start == '0'
// (Don't overdo it, however.  There's only 26 letters
//  and 10 digits to go by.)
void square(char start, int size, char (*char_provider)())
#define LOOP(i) for (i = start; i < start + size; i++)
{
  char i, j;

  printf("   "); LOOP(i) printf(" %c", i); putchar('\n');
  printf("   "); LOOP(i) printf("--"    ); putchar('\n');

  LOOP(i) {
    printf("%c |", i);
    LOOP(j) {
      printf(" %c", char_provider());
    }
    putchar('\n');
  }
}

// These strings are used to help you skew the probabilities for your
// passwords.  It can make them easier to type, while reducing their
// entropy.
// (Note that I put the space character with lower case letters.)
char* low     = " abcdefghijklmnopqrstuvwxyz";
char* low_num = " abcdefghijklmnopqrstuvwxyz0123456789";
char* up_spe =
  "!\"#$%&'()*+,-./:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`{|}~";;
char* up_spe_num =
  "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`{|}~";

// pick a random char from a string
char pick_char(char* s)
{
  return s[get_range(strlen(s))];
}


/* The following functions return any
   printable ASCII, or space. */

// Equiprobablility.
// Entropy = 6.57
char not_skewed()
{
  return get_range(95) + 32;
}

// the rare character set will appear 1/(X+1) of the time
#define MK_SKEWED(X, common, rare)    \
  char skewed_##X##common##_1##rare() \
  {                                   \
    if (get_range(X + 1) != 0)        \
      return pick_char(common);       \
    return pick_char(rare);           \
  }

// Examples of skewed probabilities, and their effect on entropy.
MK_SKEWED(1, low, up_spe_num) // Entropy = 6.42
MK_SKEWED(2, low, up_spe_num) // Entropy = 6.12
MK_SKEWED(3, low, up_spe_num) // Entropy = 5.9

// If the keyboards you intend to use have numbers in direct acces,
// you may want to use one of the following instead.
MK_SKEWED(1, low_num, up_spe)
MK_SKEWED(2, low_num, up_spe)
MK_SKEWED(3, low_num, up_spe)

int main ()
{
  //char (*cp)() = not_skewed;

  //char (*cp)() = skewed_1low_1up_spe_num;
  //char (*cp)() = skewed_2low_1up_spe_num;
  char (*cp)() = skewed_3low_1up_spe_num;

  //char (*cp)() = skewed_1low_num_1up_spe;
  //char (*cp)() = skewed_2low_num_1up_spe;
  //char (*cp)() = skewed_3low_num_1up_spe;

  square('A', 26, cp);
  //square('0', 10, cp);
  return 0;
}
